
import torch
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader
import torch.multiprocessing as mp
from .utils import round_func
import numpy as np

##################### TRANSFER ##########################

class Domain4DValues(Dataset):
    def __init__(self, quantization_interval=1):
        super(Domain4DValues, self).__init__()
        values1d = torch.arange(0, 256, quantization_interval, dtype=torch.uint8)
        values1d = torch.cat([values1d, torch.tensor([255])])
        self.quantization_interval = quantization_interval
        self.values = torch.cartesian_prod(*([values1d]*4)).view(-1, 1, 2, 2)
        
    def __getitem__(self, idx):
        if isinstance(idx, slice):
            ix1s, ix2s, ix3s, ix4s, batch = [], [], [], [], []
            for i in range(idx.start, idx.stop):
                ix1, ix2, ix3, ix4, values = self.__getitem__(i)                
                ix1s.append(ix1)
                ix2s.append(ix2)
                ix3s.append(ix3)
                ix4s.append(ix4)
                batch.append(values)
            return ix1s, ix2s, ix3s, ix4s, batch
        else:
            v = self.values[idx]
            ix = v[0]//self.quantization_interval
            return ix[0,0], ix[0,1], ix[1,0], ix[1,1], v

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        for i in range(0, len(self.values)):
            yield self.__getitem__(i)

def transfer_rc_conv(rc_conv, quantization_interval=1):
    receptive_field_pixel_count = rc_conv.window_size**2
    bucket_count = 256//quantization_interval
    lut = np.full((receptive_field_pixel_count, bucket_count+1), dtype=np.uint8, fill_value=255)
    for pixel_id in range(receptive_field_pixel_count):
        for idx, value in enumerate(range(0, 256, quantization_interval)):
            inputs = torch.tensor([value]).type(torch.float32).view(1,1,1).cuda()
            with torch.no_grad():
                outputs = rc_conv.pixel_wise_forward(inputs)            
            lut[:,idx] = outputs.flatten().cpu().numpy().astype(np.uint8)
            print(f"\r {rc_conv.__class__.__name__} {pixel_id*bucket_count + idx +1}/{receptive_field_pixel_count*bucket_count}", end="    ")
    print()
    return lut

def transfer_2x2_input_SxS_output(block, quantization_interval=16, batch_size=2**10):
    bucket_count = 256//quantization_interval    
    scale = block.upscale_factor if hasattr(block, 'upscale_factor') else 1
    lut = np.full((bucket_count+1, bucket_count+1, bucket_count+1, bucket_count+1, 1, scale, scale), dtype=np.uint8, fill_value=255) # 4DLUT for simple input window 2x2
    domain_values = Domain4DValues(quantization_interval=quantization_interval)
    domain_values_loader = DataLoader(
        domain_values, 
        batch_size=batch_size if quantization_interval >= 16 else 2**16, 
        pin_memory=True, 
        num_workers=1 if quantization_interval >= 16 else mp.cpu_count() 
    )
    counter = 0
    for idx, (ix1s, ix2s, ix3s, ix4s, batch) in enumerate(domain_values_loader):
        inputs = batch.type(torch.float32).cuda()
        with torch.no_grad():
            outputs = block(inputs)
        lut[ix1s, ix2s, ix3s, ix4s, ...] = outputs.cpu().numpy().astype(np.uint8) #[:,:,:scale,:scale] # TODO first layer automatically pad image
        counter += inputs.shape[0]
        print(f"\r {block.__class__.__name__} {counter}/{len(domain_values)}", end="    ")
    print()
    lut = lut.squeeze(-3)
    return lut



##################### FORWARD ##########################

def forward_2x2_input_SxS_output(index, lut):
    b,c,hs,ws = index.shape
    scale = lut.shape[-1]
    index = F.pad(input=index, pad=[0,1,0,1], mode='replicate') #?
    out = select_index_4dlut_tetrahedral2(
        ixA = index, 
        ixB = torch.roll(index, shifts=[0, -1], dims=[2,3]), 
        ixC = torch.roll(index, shifts=[-1, 0], dims=[2,3]), 
        ixD = torch.roll(index, shifts=[-1,-1], dims=[2,3]), 
        lut = lut
    ) 
    out = out[:,:,0:-1,0:-1,:,:] # unpad
    # Pixel Shuffle. Example: [3, 1, 126, 126, 4, 4] -> [3, 1, 126, 4, 126, 4] -> [3, 1, 504, 504]
    out = out.permute(0,1,2,4,3,5).reshape(b,1,hs*scale,ws*scale) 
    out = round_func(out)
    return out

def forward_rc_conv_centered(index, lut):
    window_size = lut.shape[0]
    index = F.pad(index, pad=[window_size//2]*4, mode='replicate') 
    window_indexes = lut.shape[:-1]         
    index = index.unsqueeze(-1)
    x = torch.zeros_like(index)
    for i in range(window_indexes[-2]):
        for j in range(window_indexes[-1]):
            shift_i, shift_j = -window_indexes[-2]//2+1 + i, -window_indexes[-1]//2+1 + j
            shifted_index = torch.roll(index, shifts=[shift_i, shift_j], dims=[-2, -1])  
            x += select_index_1dlut_linear(ixA=shifted_index, lut=lut[i,j])
    x /= window_indexes[-2]*window_indexes[-1] 
    x = x.squeeze(-1)
    x = round_func(x)
    x = x[:,:,window_size//2:-window_size//2+1,window_size//2:-window_size//2+1]
    return x

def forward_rc_conv_rot90(index, lut):
    window_size = lut.shape[0]
    index = F.pad(index, pad=[0, window_size-1]*2, mode='replicate') 
    window_indexes = lut.shape[:-1]         
    index = index.unsqueeze(-1)
    x = torch.zeros_like(index)
    for i in range(window_indexes[-2]):
        for j in range(window_indexes[-1]):
            shift_i, shift_j = i, j
            shifted_index = torch.roll(index, shifts=[shift_i, shift_j], dims=[-2, -1])  
            x += select_index_1dlut_linear(ixA=shifted_index, lut=lut[i,j])
    x /= window_indexes[-2]*window_indexes[-1] 
    x = x.squeeze(-1)
    x = round_func(x)
    x = x[:,:,:-(window_size-1),:-(window_size-1)]
    return x


##################### UTILS ##########################

def select_index_1dlut_linear(ixA, lut):
    dimA = lut.shape[0]
    qA = 256/(dimA-1)
    outDims = lut.shape[1:]
    lut = lut.reshape(dimA, *outDims).permute(*(i+1 for i in range(len(outDims))), 0)
    index_loop_indexes = ixA.shape
    lut_loop_indexes = lut.shape[:-1]    
    ixA = ixA.view(*(ixA.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    msbA = torch.floor_divide(ixA, qA).type(torch.int64)
    msbB = torch.floor_divide(ixA, qA).type(torch.int64) + 1
    lsb_index = ixA % qA
    lut = lut.view(*((1,)*len(index_loop_indexes) + lut.shape)).expand(index_loop_indexes + lut.shape)
    outA = torch.gather(input=lut, dim=-1, index=msbA) 
    outB = torch.gather(input=lut, dim=-1, index=msbB)
    out = outA + (lsb_index/qA) * (outB-outA)
    out = out.squeeze(-1)
    return out

def select_index_1dlut_msb(ixA, lut):
    dimA = lut.shape[0]
    outDims = lut.shape[1:]
    lut = lut.reshape(dimA, *outDims).permute(*(i+1 for i in range(len(outDims))), 0)
    index_loop_indexes = ixA.shape
    lut_loop_indexes = lut.shape[:-1]    
    ixA = ixA.view(*(ixA.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    msb_index = torch.floor_divide(ixA, 256/(dimA-1)).type(torch.int64) * dimA**0
    lut = lut.view(*((1,)*len(index_loop_indexes) + lut.shape)).expand(index_loop_indexes + lut.shape)
    out = torch.gather(input=lut, dim=-1, index=msb_index)
    out = out.squeeze(-1)
    return out

def select_index_4dlut_msb(ixA, ixB, ixC, ixD, lut):
    dimA, dimB, dimC, dimD = lut.shape[:4]
    qA, qB, qC, qD = 256/(dimA-1), 256/(dimB-1), 256/(dimC-1), 256/(dimD-1)
    outDims = lut.shape[4:]
    lut = lut.reshape(dimA*dimB*dimC*dimD, *outDims).permute(*(i+1 for i in range(len(outDims))), 0)
    index_loop_indexes = ixA.shape
    lut_loop_indexes = lut.shape[:-1]    
    ixA = ixA.view(*(ixA.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixB = ixB.view(*(ixB.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixC = ixC.view(*(ixC.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixD = ixD.view(*(ixD.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    msb_index = torch.floor_divide(ixA, qA) * dimA**3    
    msb_index += torch.floor_divide(ixB, qB) * dimB**2
    msb_index += torch.floor_divide(ixC, qC) * dimC**1
    msb_index += torch.floor_divide(ixD, qD) * dimD**0
    lut = lut.view(*((1,)*len(index_loop_indexes) + lut.shape)).expand(index_loop_indexes + lut.shape)
    out = torch.gather(input=lut, dim=-1, index=msb_index.type(torch.int64))
    out = out.squeeze(-1)
    return out

def select_index_4dlut_linear(ixA, ixB, ixC, ixD, lut):
    dimA, dimB, dimC, dimD = lut.shape[:4]
    qA, qB, qC, qD = 256/(dimA-1), 256/(dimB-1), 256/(dimC-1), 256/(dimD-1)
    outDims = lut.shape[4:]
    lut = lut.reshape(dimA*dimB*dimC*dimD, *outDims).permute(*(i+1 for i in range(len(outDims))), 0)
    index_loop_indexes = ixA.shape
    lut_loop_indexes = lut.shape[:-1]    
    lut = lut.view(*((1,)*len(index_loop_indexes) + lut.shape)).expand(index_loop_indexes + lut.shape)
    ixA = ixA.view(*(ixA.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixB = ixB.view(*(ixB.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixC = ixC.view(*(ixC.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixD = ixD.view(*(ixD.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    msb_index =  torch.floor_divide(ixA, qA).type(torch.int64) * dimA**3    
    msb_index += torch.floor_divide(ixB, qB).type(torch.int64) * dimB**2
    msb_index += torch.floor_divide(ixC, qC).type(torch.int64) * dimC**1
    msb_index += torch.floor_divide(ixD, qD).type(torch.int64) * dimD**0    
    outA = torch.gather(input=lut, dim=-1, index=msb_index)
    msb_index =  (torch.floor_divide(ixA, qA).type(torch.int64) + 1) * dimA**3    
    msb_index += (torch.floor_divide(ixB, qB).type(torch.int64) + 1) * dimB**2
    msb_index += (torch.floor_divide(ixC, qC).type(torch.int64) + 1) * dimC**1
    msb_index += (torch.floor_divide(ixD, qD).type(torch.int64) + 1) * dimD**0    
    outB = torch.gather(input=lut, dim=-1, index=msb_index)
    lsb_coef = ((ixA+ixB+ixC+ixD)/4 % qA) / qA
    out = outA + lsb_coef*(outB-outA)
    out = out.squeeze(-1)
    return out

def barycentric_interpolate(masks, coefs, vertices):
    i = torch.all(torch.stack(masks), dim=0, keepdim = False)
    coefs = torch.stack(coefs) * i 
    vertices = torch.stack(vertices)
    out = (coefs*vertices).sum(0)
    return i, out

def select_index_4dlut_tetrahedral(ixA, ixB, ixC, ixD, lut):
    dimA, dimB, dimC, dimD = lut.shape[:4]
    qA, qB, qC, qD = 256/(dimA-1), 256/(dimB-1), 256/(dimC-1), 256/(dimD-1)
    outDims = lut.shape[4:]
    lut = lut.reshape(dimA*dimB*dimC*dimD, *outDims).permute(*(i+1 for i in range(len(outDims))), 0)
    index_loop_indexes = ixA.shape
    lut_loop_indexes = lut.shape[:-1]    
    lut = lut.view(*((1,)*len(index_loop_indexes) + lut.shape)).expand(index_loop_indexes + lut.shape)
    ixA = ixA.view(*(ixA.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixB = ixB.view(*(ixB.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixC = ixC.view(*(ixC.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)
    ixD = ixD.view(*(ixD.shape + (1,)*len(lut_loop_indexes))).expand(index_loop_indexes + lut_loop_indexes).unsqueeze(-1)    

    msbA = torch.floor_divide(ixA, qA).type(torch.int64) 
    msbB = torch.floor_divide(ixB, qB).type(torch.int64) 
    msbC = torch.floor_divide(ixC, qC).type(torch.int64) 
    msbD = torch.floor_divide(ixD, qD).type(torch.int64) 
    fa, fb, fc, fd = ixA % qA, ixB % qB, ixC % qC, ixD % qD
    fab, fac, fad, fbc, fbd, fcd = fa>fb, fa>fc, fa>fd, fb>fc, fb>fd, fc>fd
    
    strides = torch.tensor([dimA**3, dimB**2, dimC**1, dimD**0], device=lut.device).view(-1, *((1,)*len(msbA.shape)))
    p0000 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA,   msbB,   msbC,   msbD  ])*strides).sum(0))
    p0001 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA,   msbB,   msbC,   msbD+1])*strides).sum(0))
    p0010 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA,   msbB,   msbC+1, msbD  ])*strides).sum(0))
    p0011 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA,   msbB,   msbC+1, msbD+1])*strides).sum(0))
    p0100 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA,   msbB+1, msbC,   msbD  ])*strides).sum(0))
    p0101 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA,   msbB+1, msbC,   msbD+1])*strides).sum(0))
    p0110 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA,   msbB+1, msbC+1, msbD  ])*strides).sum(0))
    p0111 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA,   msbB+1, msbC+1, msbD+1])*strides).sum(0))
    p1000 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA+1, msbB,   msbC,   msbD  ])*strides).sum(0))
    p1001 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA+1, msbB,   msbC,   msbD+1])*strides).sum(0))
    p1010 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA+1, msbB,   msbC+1, msbD  ])*strides).sum(0))
    p1011 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA+1, msbB,   msbC+1, msbD+1])*strides).sum(0))
    p1100 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA+1, msbB+1, msbC,   msbD  ])*strides).sum(0))
    p1101 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA+1, msbB+1, msbC,   msbD+1])*strides).sum(0))
    p1110 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA+1, msbB+1, msbC+1, msbD  ])*strides).sum(0))
    p1111 = torch.gather(input=lut, dim=-1, index=(torch.stack([msbA+1, msbB+1, msbC+1, msbD+1])*strides).sum(0))

    i1,   out1 = barycentric_interpolate([fab, fbc, fcd],                                            [qA-fa, fa - fb, fb - fc, fc - fd, fd], [p0000, p1000, p1100, p1110, p1111])
    i2,   out2 = barycentric_interpolate([fab, fbc, fbd,  ~(i1)],                                    [qA-fa, fa - fb, fb - fd, fd - fc, fc], [p0000, p1000, p1100, p1101, p1111])
    i3,   out3 = barycentric_interpolate([fab, fbc, fad,  ~(i1), ~(i2)],                             [qA-fa, fa - fd, fd - fb, fb - fc, fc], [p0000, p1000, p1001, p1101, p1111])
    i4,   out4 = barycentric_interpolate([fab, fbc,       ~(i1), ~(i2), ~(i3)],                      [qA-fd, fd - fa, fa - fb, fb - fc, fc], [p0000, p0001, p1001, p1101, p1111])
    i5,   out5 = barycentric_interpolate([fab, fac, fbd,  ~(fbc)],                                   [qA-fa, fa - fc, fc - fb, fb - fd, fd], [p0000, p1000, p1010, p1110, p1111])
    i6,   out6 = barycentric_interpolate([fab, fac, fcd,  ~(fbc), ~(i5)],                            [qA-fa, fa - fc, fc - fd, fd - fb, fb], [p0000, p1000, p1010, p1011, p1111])
    i7,   out7 = barycentric_interpolate([fab, fac, fad,  ~(fbc), ~(i5), ~(i6)],                     [qA-fa, fa - fd, fd - fc, fc - fb, fb], [p0000, p1000, p1001, p1011, p1111])
    i8,   out8 = barycentric_interpolate([fab, fac,       ~(fbc), ~(i5), ~(i6), ~(i7)],              [qA-fd, fd - fa, fa - fc, fc - fb, fb], [p0000, p0001, p1001, p1011, p1111])
    i9,   out9 = barycentric_interpolate([fab, fbd,       ~(fbc), ~(fac)],                           [qA-fc, fc - fa, fa - fb, fb - fd, fd], [p0000, p0010, p1010, p1110, p1111])
    i10, out10 = barycentric_interpolate([fab, fad,       ~(fbc), ~(fac), ~(i9)],                    [qA-fc, fc - fa, fa - fd, fd - fb, fb], [p0000, p0010, p1010, p1011, p1111])
    i11, out11 = barycentric_interpolate([fab, fcd,       ~(fbc), ~(fac), ~(i9), ~(i10)],            [qA-fc, fc - fd, fd - fa, fa - fb, fb], [p0000, p0010, p0011, p1011, p1111])
    i12, out12 = barycentric_interpolate([fab,            ~(fbc), ~(fac), ~(i9), ~(i10), ~(i11)],    [qA-fd, fd - fc, fc - fa, fa - fb, fb], [p0000, p0001, p0011, p1011, p1111])

    i13, out13 = barycentric_interpolate([fac, fcd, ~(fab)],                                         [qA-fb, fb - fa, fa - fc, fc - fd, fd], [p0000, p0100, p1100, p1110, p1111])
    i14, out14 = barycentric_interpolate([fac, fad, ~(fab), ~(i13)],                                 [qA-fb, fb - fa, fa - fd, fd - fc, fc], [p0000, p0100, p1100, p1101, p1111])
    i15, out15 = barycentric_interpolate([fac, fbd, ~(fab), ~(i13), ~(i14)],                         [qA-fb, fb - fd, fd - fa, fa - fc, fc], [p0000, p0100, p0101, p1101, p1111])
    i16, out16 = barycentric_interpolate([fac,      ~(fab), ~(i13), ~(i14), ~(i15) ],                [qA-fd, fd - fb, fb - fa, fa - fc, fc], [p0000, p0001, p0101, p1101, p1111])
    i17, out17 = barycentric_interpolate([fbc, fad, ~(fab), ~(fac)],                                 [qA-fb, fb - fc, fc - fa, fa - fd, fd], [p0000, p0100, p0110, p1110, p1111])
    i18, out18 = barycentric_interpolate([fbc, fcd, ~(fab), ~(fac), ~(i17)],                         [qA-fb, fb - fc, fc - fd, fd - fa, fa], [p0000, p0100, p0110, p0111, p1111])
    i19, out19 = barycentric_interpolate([fbc, fbd, ~(fab), ~(fac), ~(i17), ~(i18)],                 [qA-fb, fb - fd, fd - fc, fc - fa, fa], [p0000, p0100, p0101, p0111, p1111])
    i20, out20 = barycentric_interpolate([fbc,      ~(fab), ~(fac), ~(i17), ~(i18), ~(i19)],         [qA-fd, fd - fb, fb - fc, fc - fa, fa], [p0000, p0001, p0101, p0111, p1111])
    i21, out21 = barycentric_interpolate([fad,      ~(fab), ~(fac), ~(fbc) ],                        [qA-fc, fc - fb, fb - fa, fa - fd, fd], [p0000, p0010, p0110, p1110, p1111])
    i22, out22 = barycentric_interpolate([fbd,      ~(fab), ~(fac), ~(fbc), ~(i21)],                 [qA-fc, fc - fb, fb - fd, fd - fa, fa], [p0000, p0010, p0110, p0111, p1111])
    i23, out23 = barycentric_interpolate([fcd,      ~(fab), ~(fac), ~(fbc), ~(i21), ~(i22)],         [qA-fc, fc - fd, fd - fb, fb - fa, fa], [p0000, p0010, p0011, p0111, p1111])
    i24, out24 = barycentric_interpolate([          ~(fab), ~(fac), ~(fbc), ~(i21), ~(i22), ~(i23)], [qA-fd, fd - fc, fc - fb, fb - fa, fa], [p0000, p0001, p0011, p0111, p1111])

    out = out1 + out2 + out3 + out4 + out5 + out6 + out7 + out8 + out9 + out10 + out11 + out12 + out13 + out14 + out15 + out16 + out17 + out18 + out19 + out20 + out21 + out22 + out23 + out24
    out /= qA
    out = out.squeeze(-1)
    return out


def select_index_4dlut_tetrahedral2(ixA, ixB, ixC, ixD, lut): #self, weight, upscale, mode, img_in, bd):
    dimA, dimB, dimC, dimD = lut.shape[:4]
    q = 256/(dimA-1)
    L = dimA
    upscale = lut.shape[-1]
    weight = lut

    img_a1 = torch.floor_divide(ixA, q).type(torch.int64)
    img_b1 = torch.floor_divide(ixB, q).type(torch.int64)
    img_c1 = torch.floor_divide(ixC, q).type(torch.int64)
    img_d1 = torch.floor_divide(ixD, q).type(torch.int64)

    # Extract LSBs
    fa = ixA % q
    fb = ixB % q
    fc = ixC % q
    fd = ixD % q

    img_a2 = img_a1 + 1
    img_b2 = img_b1 + 1
    img_c2 = img_c1 + 1
    img_d2 = img_d1 + 1

    p0000 = weight[img_a1.flatten() * L * L * L + img_b1.flatten() * L * L + img_c1.flatten() * L + img_d1.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p0001 = weight[img_a1.flatten() * L * L * L + img_b1.flatten() * L * L + img_c1.flatten() * L + img_d2.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p0010 = weight[img_a1.flatten() * L * L * L + img_b1.flatten() * L * L + img_c2.flatten() * L + img_d1.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p0011 = weight[img_a1.flatten() * L * L * L + img_b1.flatten() * L * L + img_c2.flatten() * L + img_d2.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p0100 = weight[img_a1.flatten() * L * L * L + img_b2.flatten() * L * L + img_c1.flatten() * L + img_d1.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p0101 = weight[img_a1.flatten() * L * L * L + img_b2.flatten() * L * L + img_c1.flatten() * L + img_d2.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p0110 = weight[img_a1.flatten() * L * L * L + img_b2.flatten() * L * L + img_c2.flatten() * L + img_d1.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p0111 = weight[img_a1.flatten() * L * L * L + img_b2.flatten() * L * L + img_c2.flatten() * L + img_d2.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))

    p1000 = weight[img_a2.flatten() * L * L * L + img_b1.flatten() * L * L + img_c1.flatten() * L + img_d1.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p1001 = weight[img_a2.flatten() * L * L * L + img_b1.flatten() * L * L + img_c1.flatten() * L + img_d2.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p1010 = weight[img_a2.flatten() * L * L * L + img_b1.flatten() * L * L + img_c2.flatten() * L + img_d1.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p1011 = weight[img_a2.flatten() * L * L * L + img_b1.flatten() * L * L + img_c2.flatten() * L + img_d2.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p1100 = weight[img_a2.flatten() * L * L * L + img_b2.flatten() * L * L + img_c1.flatten() * L + img_d1.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p1101 = weight[img_a2.flatten() * L * L * L + img_b2.flatten() * L * L + img_c1.flatten() * L + img_d2.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p1110 = weight[img_a2.flatten() * L * L * L + img_b2.flatten() * L * L + img_c2.flatten() * L + img_d1.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    p1111 = weight[img_a2.flatten() * L * L * L + img_b2.flatten() * L * L + img_c2.flatten() * L + img_d2.flatten()].reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))

    out = torch.zeros((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale), dtype=weight.dtype).to(device=weight.device)
    sz = img_a1.shape[0] * img_a1.shape[1] * img_a1.shape[2] * img_a1.shape[3]
    out = out.reshape(sz, -1)

    p0000 = p0000.reshape(sz, -1)
    p0100 = p0100.reshape(sz, -1)
    p1000 = p1000.reshape(sz, -1)
    p1100 = p1100.reshape(sz, -1)
    fa = fa.reshape(-1, 1)

    p0001 = p0001.reshape(sz, -1)
    p0101 = p0101.reshape(sz, -1)
    p1001 = p1001.reshape(sz, -1)
    p1101 = p1101.reshape(sz, -1)
    fb = fb.reshape(-1, 1)
    fc = fc.reshape(-1, 1)

    p0010 = p0010.reshape(sz, -1)
    p0110 = p0110.reshape(sz, -1)
    p1010 = p1010.reshape(sz, -1)
    p1110 = p1110.reshape(sz, -1)
    fd = fd.reshape(-1, 1)

    p0011 = p0011.reshape(sz, -1)
    p0111 = p0111.reshape(sz, -1)
    p1011 = p1011.reshape(sz, -1)
    p1111 = p1111.reshape(sz, -1)

    fab = fa > fb;
    fac = fa > fc;
    fad = fa > fd

    fbc = fb > fc;
    fbd = fb > fd;
    fcd = fc > fd

    i1 = i = torch.all(torch.cat([fab, fbc, fcd], dim=1), dim=1);                                                        out[i] = (q - fa[i]) * p0000[i] + (fa[i] - fb[i]) * p1000[i] + (fb[i] - fc[i]) * p1100[i] + (fc[i] - fd[i]) * p1110[i] + (fd[i]) * p1111[i]
    i2 = i = torch.all(torch.cat([~i1[:, None], fab, fbc, fbd], dim=1), dim=1);                                          out[i] = (q - fa[i]) * p0000[i] + (fa[i] - fb[i]) * p1000[i] + (fb[i] - fd[i]) * p1100[i] + (fd[i] - fc[i]) * p1101[i] + (fc[i]) * p1111[i]
    i3 = i = torch.all(torch.cat([~i1[:, None], ~i2[:, None], fab, fbc, fad], dim=1), dim=1);                            out[i] = (q - fa[i]) * p0000[i] + (fa[i] - fd[i]) * p1000[i] + (fd[i] - fb[i]) * p1001[i] + (fb[i] - fc[i]) * p1101[i] + (fc[i]) * p1111[i]
    i4 = i = torch.all(torch.cat([~i1[:, None], ~i2[:, None], ~i3[:, None], fab, fbc], dim=1), dim=1);                   out[i] = (q - fd[i]) * p0000[i] + (fd[i] - fa[i]) * p0001[i] + (fa[i] - fb[i]) * p1001[i] + (fb[i] - fc[i]) * p1101[i] + (fc[i]) * p1111[i]

    i5 = i = torch.all(torch.cat([~(fbc), fab, fac, fbd], dim=1), dim=1);                                                out[i] = (q - fa[i]) * p0000[i] + (fa[i] - fc[i]) * p1000[i] + (fc[i] - fb[i]) * p1010[i] + (fb[i] - fd[i]) * p1110[i] + (fd[i]) * p1111[i]
    i6 = i = torch.all(torch.cat([~(fbc), ~i5[:, None], fab, fac, fcd], dim=1), dim=1);                                  out[i] = (q - fa[i]) * p0000[i] + (fa[i] - fc[i]) * p1000[i] + (fc[i] - fd[i]) * p1010[i] + (fd[i] - fb[i]) * p1011[i] + (fb[i]) * p1111[i]
    i7 = i = torch.all(torch.cat([~(fbc), ~i5[:, None], ~i6[:, None], fab, fac, fad], dim=1), dim=1);                    out[i] = (q - fa[i]) * p0000[i] + (fa[i] - fd[i]) * p1000[i] + (fd[i] - fc[i]) * p1001[i] + (fc[i] - fb[i]) * p1011[i] + (fb[i]) * p1111[i]
    i8 = i = torch.all(torch.cat([~(fbc), ~i5[:, None], ~i6[:, None], ~i7[:, None], fab, fac], dim=1), dim=1);           out[i] = (q - fd[i]) * p0000[i] + (fd[i] - fa[i]) * p0001[i] + (fa[i] - fc[i]) * p1001[i] + (fc[i] - fb[i]) * p1011[i] + (fb[i]) * p1111[i]

    i9 = i = torch.all(torch.cat([~(fbc), ~(fac), fab, fbd], dim=1), dim=1);                                             out[i] = (q - fc[i]) * p0000[i] + (fc[i] - fa[i]) * p0010[i] + (fa[i] - fb[i]) * p1010[i] + (fb[i] - fd[i]) * p1110[i] + (fd[i]) * p1111[i]
    # Fix the overflow bug in SR-LUT's implementation, should compare fd with fa first!
    # i10 = i = torch.all(torch.cat([~(fbc), ~(fac), ~i9[:,None], fab, fcd], dim=1), dim=1)
    # out[i] = (q-fc[i]) * p0000[i] + (fc[i]-fa[i]) * p0010[i] + (fa[i]-fd[i]) * p1010[i] + (fd[i]-fb[i]) * p1011[i] + (fb[i]) * p1111[i]
    # i11 = i = torch.all(torch.cat([~(fbc), ~(fac), ~i9[:,None], ~i10[:,None], fab, fad], dim=1), dim=1)
    # out[i] = (q-fc[i]) * p0000[i] + (fc[i]-fd[i]) * p0010[i] + (fd[i]-fa[i]) * p0011[i] + (fa[i]-fb[i]) * p1011[i] + (fb[i]) * p1111[i]
    # c > a > d > b
    i10 = i = torch.all(torch.cat([~(fbc), ~(fac), ~i9[:, None], fab, fad], dim=1), dim=1);                              out[i] = (q - fc[i]) * p0000[i] + (fc[i] - fa[i]) * p0010[i] + (fa[i] - fd[i]) * p1010[i] + (fd[i] - fb[i]) * p1011[i] + (fb[i]) * p1111[i]
    # c > d > a > b
    i11 = i = torch.all(torch.cat([~(fbc), ~(fac), ~i9[:, None], ~i10[:, None], fab, fcd], dim=1), dim=1);               out[i] = (q - fc[i]) * p0000[i] + (fc[i] - fd[i]) * p0010[i] + (fd[i] - fa[i]) * p0011[i] + (fa[i] - fb[i]) * p1011[i] + (fb[i]) * p1111[i]
    i12 = i = torch.all(torch.cat([~(fbc), ~(fac), ~i9[:, None], ~i10[:, None], ~i11[:, None], fab], dim=1), dim=1);     out[i] = (q - fd[i]) * p0000[i] + (fd[i] - fc[i]) * p0001[i] + (fc[i] - fa[i]) * p0011[i] + (fa[i] - fb[i]) * p1011[i] + (fb[i]) * p1111[i]

    i13 = i = torch.all(torch.cat([~(fab), fac, fcd], dim=1), dim=1);                                                    out[i] = (q - fb[i]) * p0000[i] + (fb[i] - fa[i]) * p0100[i] + (fa[i] - fc[i]) * p1100[i] + (fc[i] - fd[i]) * p1110[i] + (fd[i]) * p1111[i]
    i14 = i = torch.all(torch.cat([~(fab), ~i13[:, None], fac, fad], dim=1), dim=1);                                     out[i] = (q - fb[i]) * p0000[i] + (fb[i] - fa[i]) * p0100[i] + (fa[i] - fd[i]) * p1100[i] + (fd[i] - fc[i]) * p1101[i] + (fc[i]) * p1111[i]
    i15 = i = torch.all(torch.cat([~(fab), ~i13[:, None], ~i14[:, None], fac, fbd], dim=1), dim=1);                      out[i] = (q - fb[i]) * p0000[i] + (fb[i] - fd[i]) * p0100[i] + (fd[i] - fa[i]) * p0101[i] + (fa[i] - fc[i]) * p1101[i] + (fc[i]) * p1111[i]
    i16 = i = torch.all(torch.cat([~(fab), ~i13[:, None], ~i14[:, None], ~i15[:, None], fac], dim=1), dim=1);            out[i] = (q - fd[i]) * p0000[i] + (fd[i] - fb[i]) * p0001[i] + (fb[i] - fa[i]) * p0101[i] + (fa[i] - fc[i]) * p1101[i] + (fc[i]) * p1111[i]

    i17 = i = torch.all(torch.cat([~(fab), ~(fac), fbc, fad], dim=1), dim=1);                                            out[i] = (q - fb[i]) * p0000[i] + (fb[i] - fc[i]) * p0100[i] + (fc[i] - fa[i]) * p0110[i] + (fa[i] - fd[i]) * p1110[i] + (fd[i]) * p1111[i]
    i18 = i = torch.all(torch.cat([~(fab), ~(fac), ~i17[:, None], fbc, fcd], dim=1), dim=1);                             out[i] = (q - fb[i]) * p0000[i] + (fb[i] - fc[i]) * p0100[i] + (fc[i] - fd[i]) * p0110[i] + (fd[i] - fa[i]) * p0111[i] + (fa[i]) * p1111[i]
    i19 = i = torch.all(torch.cat([~(fab), ~(fac), ~i17[:, None], ~i18[:, None], fbc, fbd], dim=1), dim=1);              out[i] = (q - fb[i]) * p0000[i] + (fb[i] - fd[i]) * p0100[i] + (fd[i] - fc[i]) * p0101[i] + (fc[i] - fa[i]) * p0111[i] + (fa[i]) * p1111[i]
    i20 = i = torch.all(torch.cat([~(fab), ~(fac), ~i17[:, None], ~i18[:, None], ~i19[:, None], fbc], dim=1), dim=1);    out[i] = (q - fd[i]) * p0000[i] + (fd[i] - fb[i]) * p0001[i] + (fb[i] - fc[i]) * p0101[i] + (fc[i] - fa[i]) * p0111[i] + (fa[i]) * p1111[i]

    i21 = i = torch.all(torch.cat([~(fab), ~(fac), ~(fbc), fad], dim=1), dim=1);                                         out[i] = (q - fc[i]) * p0000[i] + (fc[i] - fb[i]) * p0010[i] + (fb[i] - fa[i]) * p0110[i] + (fa[i] - fd[i]) * p1110[i] + (fd[i]) * p1111[i]
    i22 = i = torch.all(torch.cat([~(fab), ~(fac), ~(fbc), ~i21[:, None], fbd], dim=1), dim=1);                          out[i] = (q - fc[i]) * p0000[i] + (fc[i] - fb[i]) * p0010[i] + (fb[i] - fd[i]) * p0110[i] + (fd[i] - fa[i]) * p0111[i] + (fa[i]) * p1111[i]
    i23 = i = torch.all(torch.cat([~(fab), ~(fac), ~(fbc), ~i21[:, None], ~i22[:, None], fcd], dim=1), dim=1);           out[i] = (q - fc[i]) * p0000[i] + (fc[i] - fd[i]) * p0010[i] + (fd[i] - fb[i]) * p0011[i] + (fb[i] - fa[i]) * p0111[i] + (fa[i]) * p1111[i]
    i24 = i = torch.all(torch.cat([~(fab), ~(fac), ~(fbc), ~i21[:, None], ~i22[:, None], ~i23[:, None]], dim=1), dim=1); out[i] = (q - fd[i]) * p0000[i] + (fd[i] - fc[i]) * p0001[i] + (fc[i] - fb[i]) * p0011[i] + (fb[i] - fa[i]) * p0111[i] + (fa[i]) * p1111[i]

    out = out.reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2], img_a1.shape[3], upscale, upscale))
    out = out.permute(0, 1, 2, 4, 3, 5).reshape((img_a1.shape[0], img_a1.shape[1], img_a1.shape[2] * upscale, img_a1.shape[3] * upscale))
    out = out / q
    return out